import os
import json
import codecs
import datetime
from zipfile import ZipFile
from pathlib import PurePath, Path

from bofhound.logger import logger
from bofhound import console
from bofhound.ad.models import BloodHoundDomain, BloodHoundComputer, BloodHoundUser, BloodHoundGroup, BloodHoundSchema, BloodHoundEnterpriseCA, BloodHoundAIACA, BloodHoundRootCA, BloodHoundCertTemplate, BloodHoundContainer

class BloodHoundWriter():
    files = []
    ct = None

    @staticmethod
    def write(out_dir='.', domains=None, computers=None, users=None,
          groups=None, ous=None, containers=None, gpos=None, enterprisecas=None, aiacas=None,
          rootcas=None, ntauthstores=None, issuancepolicies=None, certtemplates=None, 
          trusts=None, trustaccounts=None, properties_level=2, zip_files=False):

        os.makedirs(out_dir, exist_ok=True)
        BloodHoundWriter.ct = BloodHoundWriter.timestamp()

        outfiles = []

        if domains is not None:
            # print(BloodHoundSchema.ObjectTypeGuidMap)
            with console.status(" [bold] Writing domains to JSON...\n", spinner="aesthetic"):
                outfiles.append(
                    BloodHoundWriter.write_domain_file(out_dir, domains, properties_level)
                )

        if computers is not None:
            with console.status(" [bold] Writing computers to JSON...\n", spinner="aesthetic"):
                outfiles.append(
                    BloodHoundWriter.write_computers_file(out_dir, computers, properties_level)
                )

        if users is not None:
            with console.status(" [bold] Writing users to JSON...\n", spinner="aesthetic"):
                outfiles.append(
                    BloodHoundWriter.write_users_file(out_dir, users, properties_level)
                )

        if groups is not None:
            with console.status(" [bold] Writing groups to JSON...\n", spinner="aesthetic"):
                outfiles.append(
                    BloodHoundWriter.write_groups_file(out_dir, groups, properties_level)
                )

        if ous is not None:
            with console.status(" [bold] Writing OUs to JSON...\n", spinner="aesthetic"):
                outfiles.append(
                    BloodHoundWriter.write_ous_file(out_dir, ous, properties_level)
                )
        
        if containers is not None:
            with console.status(" [bold] Writing Containers to JSON...\n", spinner="aesthetic"):
                outfiles.append(
                    BloodHoundWriter.write_containers_file(out_dir, containers, properties_level)
                )

        if gpos is not None:
            with console.status(" [bold] Writing GPOs to JSON...\n", spinner="aesthetic"):
                outfiles.append(
                    BloodHoundWriter.write_gpos_file(out_dir, gpos, properties_level)
                )

        if enterprisecas is not None:
            with console.status(" [bold] Writing Enterprise CAs to JSON...\n", spinner="aesthetic"):
                outfiles.append(
                    BloodHoundWriter.write_enterprisecas_file(out_dir, enterprisecas, properties_level)
                )

        if aiacas is not None:
            with console.status(" [bold] Writing AIA CAs to JSON...\n", spinner="aesthetic"):
                outfiles.append(
                    BloodHoundWriter.write_aiacas_file(out_dir, aiacas, properties_level)
                )

        if rootcas is not None:
            with console.status(" [bold] Writing Root CAs to JSON...\n", spinner="aesthetic"):
                outfiles.append(
                    BloodHoundWriter.write_rootcas_file(out_dir, rootcas, properties_level)
                )

        if ntauthstores is not None:
            with console.status(" [bold] Writing NTAuth Stores to JSON...\n", spinner="aesthetic"):
                outfiles.append(
                    BloodHoundWriter.write_ntauthstores_file(out_dir, ntauthstores, properties_level)
                )

        if issuancepolicies is not None:
            with console.status(" [bold] Writing Issuance Policies to JSON...\n", spinner="aesthetic"):
                outfiles.append(
                    BloodHoundWriter.write_issuancepolicies_file(out_dir, issuancepolicies, properties_level)
                )
        
        if certtemplates is not None:
            with console.status(" [bold] Writing Cert Templates to JSON...\n", spinner="aesthetic"):
                outfiles.append(
                    BloodHoundWriter.write_certtemplates_file(out_dir, certtemplates, properties_level)
                )

        if trusts is not None:
            outfiles.append(
                BloodHoundWriter.write_trusts_file(out_dir, trusts, properties_level)
            )

        if trustaccounts is not None:
            outfiles.append(
                BloodHoundWriter.write_trustaccounts_file(out_dir, trustaccounts, properties_level)
            )

        if out_dir == ".":
            logger.info(f'JSON files written to current directory')
        else:
            logger.info(f'JSON files written to {out_dir}')

        if zip_files:
            zip_name = PurePath(out_dir, f"bloodhound_{BloodHoundWriter.ct}.zip")
            with ZipFile(zip_name, "w") as zip:
                for bh_file in BloodHoundWriter.files:
                    zip.write(bh_file, bh_file.name)
                    Path(bh_file).unlink()

            #
            # Single zipfile can be uploaded instad of all JSON files
            #  override outfiles list with the zip file
            #
            outfiles = [zip_name]

            logger.info(f'Files compressed into {zip_name}')

        # remove any 'None' entries from the outfiles list
        return [f for f in outfiles if f is not None]
    

    @staticmethod
    def write_domain_file(out_dir, domains, properties_level):
        if len(domains) == 0:
            return

        datastruct = {
            "data": [],
            "meta": {
                "type": "domains",
                "count": 0,
                "methods": 0,
                "version": 5
            }
        }

        for domain in domains:
            datastruct['data'].append(domain.to_json(properties_level))
            datastruct['meta']['count'] += 1

        out_file = PurePath(out_dir, f'domains_{BloodHoundWriter.ct}.json')
        BloodHoundWriter.files.append(out_file)
        with codecs.open(out_file, 'w', 'utf-8') as f:
            json.dump(datastruct, f, ensure_ascii=False)
        
        return out_file
    

    @staticmethod
    def write_computers_file(out_dir, computers, properties_level):
        if len(computers) == 0:
            return

        datastruct = {
            "data": [],
            "meta": {
                "type": "computers",
                "count": 0,
                "methods": 0,
                "version": 6
            }
        }

        for computer in computers:
            datastruct['data'].append(computer.to_json(properties_level))
            datastruct['meta']['count'] += 1

        out_file = PurePath(out_dir, f'computers_{BloodHoundWriter.ct}.json')
        BloodHoundWriter.files.append(out_file)
        with codecs.open(out_file, 'w', 'utf-8') as f:
            json.dump(datastruct, f, ensure_ascii=False)

        return out_file


    @staticmethod
    def write_users_file(out_dir, users, properties_level):
        if len(users) == 0:
            return

        datastruct = {
            "data": [],
            "meta": {
                "type": "users",
                "count": 0,
                "methods": 0,
                "version": 6
            }
        }

        for user in users:
            datastruct['data'].append(user.to_json(properties_level))
            datastruct['meta']['count'] += 1

        out_file = PurePath(out_dir, f'users_{BloodHoundWriter.ct}.json')
        BloodHoundWriter.files.append(out_file)
        with codecs.open(out_file, 'w', 'utf-8') as f:
            json.dump(datastruct, f, ensure_ascii=False)

        return out_file


    @staticmethod
    def write_groups_file(out_dir, groups, properties_level):
        if len(groups) == 0:
            return

        datastruct = {
            "data": [],
            "meta": {
                "type": "groups",
                "count": 0,
                "methods": 0,
                "version": 6
            }
        }

        for group in groups:
            datastruct['data'].append(group.to_json(properties_level))
            datastruct['meta']['count'] += 1

        out_file = PurePath(out_dir, f'groups_{BloodHoundWriter.ct}.json')
        BloodHoundWriter.files.append(out_file)
        with codecs.open(out_file, 'w', 'utf-8') as f:
            json.dump(datastruct, f, ensure_ascii=False)

        return out_file

    
    @staticmethod
    def write_ous_file(out_dir, ous, properties_level):
        if len(ous) == 0:
            return

        datastruct = {
            "data": [],
            "meta": {
                "type": "ous",
                "count": 0,
                "methods": 0,
                "version": 6
            }
        }

        for ou in ous:
            datastruct['data'].append(ou.to_json(properties_level))
            datastruct['meta']['count'] += 1

        out_file = PurePath(out_dir, f'ous_{BloodHoundWriter.ct}.json')
        BloodHoundWriter.files.append(out_file)
        with codecs.open(out_file, 'w', 'utf-8') as f:
            json.dump(datastruct, f, ensure_ascii=False)

        return out_file


    @staticmethod
    def write_containers_file(out_dir, containers, properties_level):
        if len(containers) == 0:
            return

        datastruct = {
            "data": [],
            "meta": {
                "type": "containers",
                "count": 0,
                "methods": 0,
                "version": 6
            }
        }

        for container in containers:
            datastruct['data'].append(container.to_json(properties_level))
            datastruct['meta']['count'] += 1

        out_file = PurePath(out_dir, f'containers_{BloodHoundWriter.ct}.json')
        BloodHoundWriter.files.append(out_file)
        with codecs.open(out_file, 'w', 'utf-8') as f:
            json.dump(datastruct, f, ensure_ascii=False)

        return out_file


    @staticmethod
    def write_gpos_file(out_dir, gpos, properties_level):
        if len(gpos) == 0:
            return

        datastruct = {
            "data": [],
            "meta": {
                "type": "gpos",
                "count": 0,
                "methods": 0,
                "version": 5
            }
        }

        for gpo in gpos:
            datastruct['data'].append(gpo.to_json(properties_level))
            datastruct['meta']['count'] += 1

        out_file = PurePath(out_dir, f'gpos_{BloodHoundWriter.ct}.json')
        BloodHoundWriter.files.append(out_file)
        with codecs.open(out_file, 'w', 'utf-8') as f:
            json.dump(datastruct, f, ensure_ascii=False)

        return out_file
    

    @staticmethod
    def write_enterprisecas_file(out_dir, enterprisecas, properties_level):
        if len(enterprisecas) == 0:
            return

        datastruct = {
            "data": [],
            "meta": {
                "methods" : 0,
                "type": "enterprisecas",
                "count": 0,
                "version": 6
            }
        }

        for enterpriseca in enterprisecas:
            datastruct['data'].append(enterpriseca.to_json(properties_level))
            datastruct['meta']['count'] += 1

        out_file = PurePath(out_dir, f'enterprisecas_{BloodHoundWriter.ct}.json')
        BloodHoundWriter.files.append(out_file)
        with codecs.open(out_file, 'w', 'utf-8') as f:
            json.dump(datastruct, f, ensure_ascii=False)

        return out_file


    @staticmethod
    def write_aiacas_file(out_dir, aiacas, properties_level):
        if len(aiacas) == 0:
            return

        datastruct = {
            "data": [],
            "meta": {
                "methods" : 0,
                "type": "aiacas",
                "count": 0,
                "version": 6
            }
        }

        for aiaca in aiacas:
            datastruct['data'].append(aiaca.to_json(properties_level))
            datastruct['meta']['count'] += 1

        out_file = PurePath(out_dir, f'aiacas_{BloodHoundWriter.ct}.json')
        BloodHoundWriter.files.append(out_file)
        with codecs.open(out_file, 'w', 'utf-8') as f:
            json.dump(datastruct, f, ensure_ascii=False)

        return out_file

    
    @staticmethod
    def write_rootcas_file(out_dir, rootcas, properties_level):
        if len(rootcas) == 0:
            return

        datastruct = {
            "data": [],
            "meta": {
                "methods" : 0,
                "type": "rootcas",
                "count": 0,
                "version": 6
            }
        }

        for rootca in rootcas:
            datastruct['data'].append(rootca.to_json(properties_level))
            datastruct['meta']['count'] += 1

        out_file = PurePath(out_dir, f'rootcas_{BloodHoundWriter.ct}.json')
        BloodHoundWriter.files.append(out_file)
        with codecs.open(out_file, 'w', 'utf-8') as f:
            json.dump(datastruct, f, ensure_ascii=False)

        return out_file

    
    @staticmethod
    def write_ntauthstores_file(out_dir, ntauthstores, properties_level):
        if len(ntauthstores) == 0:
            return

        datastruct = {
            "data": [],
            "meta": {
                "methods" : 0,
                "type": "ntauthstores",
                "count": 0,
                "version": 6
            }
        }

        for ntauthstore in ntauthstores:
            datastruct['data'].append(ntauthstore.to_json(properties_level))
            datastruct['meta']['count'] += 1

        out_file = PurePath(out_dir, f'ntauthstores_{BloodHoundWriter.ct}.json')
        BloodHoundWriter.files.append(out_file)
        with codecs.open(out_file, 'w', 'utf-8') as f:
            json.dump(datastruct, f, ensure_ascii=False)

        return out_file


    @staticmethod
    def write_issuancepolicies_file(out_dir, issuancepolicies, properties_level):
        if len(issuancepolicies) == 0:
            return

        datastruct = {
            "data": [],
            "meta": {
                "methods" : 0,
                "type": "issuancepolicies",
                "count": 0,
                "version": 6
            }
        }

        for issuancepolicy in issuancepolicies:
            datastruct['data'].append(issuancepolicy.to_json(properties_level))
            datastruct['meta']['count'] += 1

        out_file = PurePath(out_dir, f'issuancepolicies_{BloodHoundWriter.ct}.json')
        BloodHoundWriter.files.append(out_file)
        with codecs.open(out_file, 'w', 'utf-8') as f:
            json.dump(datastruct, f, ensure_ascii=False)

        return out_file

    
    @staticmethod
    def write_certtemplates_file(out_dir, certtemplates, properties_level):
        if len(certtemplates) == 0:
            return

        datastruct = {
            "data": [],
            "meta": {
                "methods": 0,
                "type": "certtemplates",
                "count": 0,
                "version": 6
            }
        }

        for certtemplate in certtemplates:
            datastruct['data'].append(certtemplate.to_json(properties_level))
            datastruct['meta']['count'] += 1

        out_file = PurePath(out_dir, f'certtemplates_{BloodHoundWriter.ct}.json')
        BloodHoundWriter.files.append(out_file)
        with codecs.open(out_file, 'w', 'utf-8') as f:
            json.dump(datastruct, f, ensure_ascii=False)

        return out_file


    @staticmethod
    def write_trusts_file(out_dir, trusts, properties_level):
        pass


    @staticmethod
    def write_trustaccounts_file(out_dir, trustaccounts, properties_level):
        pass

    @staticmethod
    def timestamp():
        ct = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        return ct



from bloodhound.ad.utils import ADUtils
from .bloodhound_object import BloodHoundObject
from bofhound.logger import logger, OBJ_EXTRA_FMT, ColorScheme


class BloodHoundAIACA(BloodHoundObject):

    GUI_PROPERTIES = [
        'domain', 'name', 'distinguishedname', 'domainsid', 'isaclprotected',
        'description', 'whencreated', 'crosscertificatepair', 'hascrosscertificatepair',
        'certthumbprint', 'certname', 'certchain', 'hasbasicconstraints',
        'basicconstraintpathlength'
    ]

    COMMON_PROPERTIES = [
    ]

    def __init__(self, object):
        super().__init__(object)

        self._entry_type = "AIACA"
        self.ContainedBy = {}
        self.IsACLProtected = False
        self.IsDeleted = False
        self.x509Certificate = None

        if 'objectguid' in object.keys():
            self.ObjectIdentifier = object.get("objectguid").upper()

        if 'distinguishedname' in object.keys():
            domain = ADUtils.ldap2domain(object.get('distinguishedname')).upper()
            self.Properties['domain'] = domain
            self.Properties['distinguishedname'] = object.get('distinguishedname').upper()

        if 'description' in object.keys():
            self.Properties['description'] = object.get('description')
        else:
            self.Properties['description'] = None

        if 'name' in object.keys():
            if 'domain' in self.Properties.keys():
                self.Properties['name'] = object.get('name').upper() + "@" + self.Properties['domain'].upper()

        if 'crosscertificatepair' in object.keys():
            self.Properties['crosscertificatepair'] = object.get('crosscertificatepair')
            self.Properties['hascrosscertificatepair'] = True
        else:
            self.Properties['crosscertificatepair'] = []
            self.Properties['hascrosscertificatepair'] = False

        if 'cacertificate' in object.keys():
            self.parse_cacertificate(object)
        
        if 'ntsecuritydescriptor' in object.keys():
            self.RawAces = object['ntsecuritydescriptor']

        
    def to_json(self, properties_level):
        self.Properties['isaclprotected'] = self.IsACLProtected
        data = super().to_json(properties_level)

        data["Aces"] = self.Aces
        data["IsDeleted"] = self.IsDeleted
        data["IsACLProtected"] = self.IsACLProtected
        data["ObjectIdentifier"] = self.ObjectIdentifier
        data["ContainedBy"] = self.ContainedBy

        return data



from .sharedparser import SharedParser


class RegSessionBofParser(SharedParser):
    START_BOUNDARY  = "-----------Registry Session---------"
    END_BOUNDARY    = "---------End Registry Session-------"
    OBJECT_TYPE     = "RegistrySession"

    def __init__(self):
        pass


    @staticmethod
    def is_start_boundary_line(line):
        return line.strip() == RegSessionBofParser.START_BOUNDARY
    

    @staticmethod
    def is_end_boundary_line(line):
        return line.strip() == RegSessionBofParser.END_BOUNDARY


from bloodhound.ad.utils import ADUtils

from .bloodhound_object import BloodHoundObject
from bofhound.logger import logger, OBJ_EXTRA_FMT, ColorScheme


class BloodHoundGPO(BloodHoundObject):

    GUI_PROPERTIES = [
        'distinguishedname', 'whencreated',
        'domain', 'domainsid', 'name', 'highvalue',
        'description', 'gpcpath', 'isaclprotected'
    ]

    COMMON_PROPERTIES = [
    ]

    def __init__(self, object):
        super().__init__(object)

        self._entry_type = "GPO"
        self.ContainedBy = {}
        
        if 'distinguishedname' in object.keys() and 'displayname' in object.keys():
            self.Properties["domain"] = ADUtils.ldap2domain(object.get('distinguishedname').upper())
            self.Properties["name"] = f"{object.get('displayname').upper()}@{self.Properties['domain']}"
            logger.debug(f"Reading GPO object {ColorScheme.gpo}{self.Properties['name']}[/]", extra=OBJ_EXTRA_FMT)

        if 'objectguid' in object.keys():
            self.ObjectIdentifier = object.get("objectguid").upper().upper()

        if 'ntsecuritydescriptor' in object.keys():
            self.RawAces = object['ntsecuritydescriptor']

        if 'description' in object.keys():
            self.Properties["description"] = object.get('description')

        if 'gpcfilesyspath' in object.keys():
            self.Properties["gpcpath"] = object.get('gpcfilesyspath')

        self.Properties["highvalue"] = False

        self.Aces = []

        self.IsDeleted = False
        self.IsACLProtected = False

    def to_json(self, properties_level):
        self.Properties['isaclprotected'] = self.IsACLProtected
        gpo = super().to_json(properties_level)

        gpo["ObjectIdentifier"] = self.ObjectIdentifier
        gpo["ContainedBy"] = self.ContainedBy
        # The below is all unsupported as of now.
        gpo["Aces"] = self.Aces
        gpo["IsDeleted"] = self.IsDeleted
        gpo["IsACLProtected"] = self.IsACLProtected

        return gpo



import base64
from uuid import UUID

from bofhound.logger import logger, OBJ_EXTRA_FMT, ColorScheme


class BloodHoundSchema(object):

	def __init__(self, object):
		self.Name = None
		self.SchemaIdGuid = None

		if 'name' in object.keys() and 'schemaidguid' in object.keys():
			self.Name = object.get('name').lower()
			try:
				value = object.get('schemaidguid')
				if '-' in value:
					self.SchemaIdGuid = value.lower()
				else:
					self.SchemaIdGuid = str(UUID(bytes_le=base64.b64decode(value))).lower()
				logger.debug(f"Reading Schema object {ColorScheme.schema}{self.Name}[/]", extra=OBJ_EXTRA_FMT)
			except:
				logger.warning(f"Error base64 decoding SchemaIDGUID attribute on Schema {ColorScheme.schema}{self.Name}[/]", extra=OBJ_EXTRA_FMT)



from .sharedparser import SharedParser


class NetLoggedOnBofParser(SharedParser):
    START_BOUNDARY  = "-----------Logged on User-----------"
    END_BOUNDARY    = "---------End Logged on User---------"
    OBJECT_TYPE     = "PrivilegedSession"

    def __init__(self):
        pass


    @staticmethod
    def is_start_boundary_line(line):
        return line.strip() == NetLoggedOnBofParser.START_BOUNDARY
    

    @staticmethod
    def is_end_boundary_line(line):
        return line.strip() == NetLoggedOnBofParser.END_BOUNDARY


from bloodhound.ad.utils import ADUtils
from bloodhound.ad.structures import LDAP_SID
from bloodhound.enumeration.memberships import MembershipEnumerator

from .bloodhound_object import BloodHoundObject
from bofhound.logger import logger, OBJ_EXTRA_FMT, ColorScheme


class BloodHoundUser(BloodHoundObject):

    GUI_PROPERTIES = [
        'domain', 'name', 'distinguishedname', 'domainsid', 'samaccountname',
        'isaclprotected', 'description', 'whencreated', 'sensitive',
        'dontreqpreauth', 'passwordnotreqd', 'unconstraineddelegation',
        'pwdneverexpires', 'enabled', 'trustedtoauth', 'lastlogon', 'lastlogontimestamp',
        'pwdlastset', 'serviceprincipalnames', 'hasspn', 'admincount',
        'displayname', 'email', 'title', 'homedirectory', 'userpassword',
        'unixpassword', 'unicodepassword', 'sfupassword', 'logonscript', 'sidhistory'
    ]

    COMMON_PROPERTIES = [
        'samaccounttype', 'objectsid', 'primarygroupid', 'isdeleted',
        'msds-groupmsamembership', 'serviceprincipalname', 'useraccountcontrol', 
        'mail', 'msds-allowedtodelegateto', 'unicodepwd', 'allowedtodelegate',   
        'memberof'
    ]

    def __init__(self, object=None):
        super().__init__(object)

        self._entry_type = "User"
        self.PrimaryGroupSid = None
        self.AllowedToDelegate = []
        self.Aces = []
        self.ContainedBy = {}
        self.SPNTargets = []
        self.HasSIDHistory = []
        self.IsACLProtected = False
        self.MemberOfDNs = []

        if isinstance(object, dict):
            self.PrimaryGroupSid = self.get_primary_membership(object) # Returns none if not exist

            if 'distinguishedname' in object.keys() and 'samaccountname' in object.keys():
                domain = ADUtils.ldap2domain(object.get('distinguishedname')).upper()
                name = f'{object.get("samaccountname")}@{domain}'.upper()
                self.Properties["name"] = name
                self.Properties["domain"] = domain
                logger.debug(f"Reading User object {ColorScheme.user}{name}[/]", extra=OBJ_EXTRA_FMT)

            if 'admincount' in object.keys():
                self.Properties["admincount"] = int(object.get('admincount')) == 1 # do not move this lower, it may break imports for users

            # self.Properties["highvalue"] = False,

            if 'useraccountcontrol' in object.keys():
                uac = int(object.get('useraccountcontrol', 0))
                self.Properties["unconstraineddelegation"] = uac & 0x00080000 == 0x00080000
                self.Properties["passwordnotreqd"] = uac & 0x00000020 == 0x00000020
                self.Properties["enabled"] = uac & 2 == 0
                self.Properties["dontreqpreauth"] = uac & 0x00400000 == 0x00400000
                self.Properties["sensitive"] = uac & 0x00100000 == 0x00100000
                self.Properties["trustedtoauth"] = uac & 0x01000000 == 0x01000000
                self.Properties["pwdneverexpires"] = uac & 0x00010000 == 0x00010000

            if 'lastlogon' in object.keys():
                self.Properties["lastlogon"] = ADUtils.win_timestamp_to_unix(
                    int(object.get('lastlogon'))
                )

            if 'lastlogontimestamp' in object.keys():
                self.Properties["lastlogontimestamp"] = ADUtils.win_timestamp_to_unix(
                    int(object.get('lastlogontimestamp'))
                )

            if 'pwdlastset' in object.keys():
                self.Properties["pwdlastset"] = ADUtils.win_timestamp_to_unix(
                    int(object.get('pwdlastset'))
                )

            if 'useraccountcontrol' in object.keys():
                self.Properties["dontreqpreauth"] = int(object.get('useraccountcontrol', 0)) & 0x00400000 == 0x00400000
                self.Properties["pwdneverexpires"] = int(object.get('useraccountcontrol', 0)) & 0x00010000 == 0x00010000
                self.Properties["sensitive"] = int(object.get('useraccountcontrol', 0)) & 0x00100000 == 0x00100000

            if 'serviceprincipalname' in object.keys():
                self.Properties["serviceprincipalnames"] = object.get('serviceprincipalname').split(',')
                self.Properties['hasspn'] = True
            else:
                self.Properties["serviceprincipalnames"] = []
                self.Properties['hasspn'] = False

            if 'serviceprincipalname' in object.keys():
                self.Properties["hasspn"] = len(object.get('serviceprincipalname', [])) > 0

            if 'samaccounttype' in object.keys():
                self.Properties["samaccounttype"] = object.get('samaccounttype')
            
            if 'displayname' in object.keys():
                self.Properties["displayname"] = object.get('displayname')

            if 'mail' in object.keys():
                self.Properties["email"] = object.get('mail')

            if 'title' in object.keys():
                self.Properties["title"] = object.get('title')

            if 'homedirectory' in object.keys():
                self.Properties["homedirectory"] = object.get('homedirectory')

            if 'description' in object.keys():
                self.Properties["description"] = object.get('description')

            if 'userpassword' in object.keys():
                self.Properties["userpassword"] = ADUtils.ensure_string(object.get('userpassword'))

            if 'sidhistory' in object.keys():
                self.Properties["sidhistory"] = [LDAP_SID(bsid).formatCanonical() for bsid in object.get('sIDHistory', [])]

            if 'msds-allowedtodelegateto' in object.keys():
                if len(object.get('msds-allowedtodelegateto', [])) > 0:
                    self.Properties['allowedtodelegate'] = object.get('msds-allowedtodelegateto', [])

            if 'ntsecuritydescriptor' in object.keys():
                self.RawAces = object['ntsecuritydescriptor']

            if 'memberof' in object.keys():
                self.MemberOfDNs = [f'CN={dn.upper()}' for dn in object.get('memberof').split(', CN=')]
                if len(self.MemberOfDNs) > 0:
                    self.MemberOfDNs[0] = self.MemberOfDNs[0][3:]

            ### TODO Not supported for the moment
            # self.Properties['displayname'] = None
            # self.Properties['email'] = None
            # self.Properties['title'] = None
            # self.Properties['homedirectory'] = None
            # self.Properties['userpassword'] = None
            # self.Properties['unixpassword'] = None
            # self.Properties['unicodepassword'] = None
            # self.Properties['sfupassword'] = None
            # self.Properties['logonscript'] = None
            # self.Properties['sidhistory'] = []


    def to_json(self, properties_level):
        self.Properties['isaclprotected'] = self.IsACLProtected
        user = super().to_json(properties_level)

        user["ObjectIdentifier"] = self.ObjectIdentifier
        user["ContainedBy"] = self.ContainedBy
        user["AllowedToDelegate"] = self.AllowedToDelegate
        user["PrimaryGroupSID"] = self.PrimaryGroupSid

        user["Aces"] = self.Aces
        user["SPNTargets"] = self.SPNTargets
        user["HasSIDHistory"] = self.HasSIDHistory
        user["IsACLProtected"] = self.IsACLProtected

        # TODO: RBCD
        # Process resource-based constrained delegation
        # _, aces = parse_binary_acl(data,
        #                            'computer',
        #                            object.get('msDS-AllowedToActOnBehalfOfOtherIdentity'),
        #                            self.addc.objecttype_guid_map)
        # outdata = self.aceresolver.resolve_aces(aces)
        # for delegated in outdata:
        #     if delegated['RightName'] == 'Owner':
        #         continue
        #     if delegated['RightName'] == 'GenericAll':
        #         data['AllowedToAct'].append({'MemberId': delegated['PrincipalSID'], 'MemberType': delegated['PrincipalType']})
        #
        # # Run ACL collection if this was not already done centrally
        # if 'acl' in collect and not skip_acl:
        #     _, aces = parse_binary_acl(data,
        #                                'computer',
        #                                object.get('nTSecurityDescriptor',
        #                                                           raw=True),
        #                                self.addc.objecttype_guid_map)
        #     # Parse aces
        #     data['Aces'] = self.aceresolver.resolve_aces(aces)

        return user



from .ldap_search_bof import LdapSearchBofParser
from .brc4_ldap_sentinel import Brc4LdapSentinelParser
from .havoc import HavocParser
from .parsertype import ParserType
from .outflankc2 import OutflankC2JsonParser
from .mythic import MythicParser


from bofhound.logger import logger, ColorScheme, OBJ_EXTRA_FMT


class LocalSession:
    SESSION_PTR             = "PTR"
    SESSION_USER            = "User"
    SESSION_COMPUTER_NAME   = "ComputerName"
    SESSION_COMPUTER_DOMAIN = "ComputerDomain"

    def __init__(self, object):
        self.username                   = None
        self.ptr_record                 = None
        self.computer_name              = None
        self.computer_netbios_domain    = None
        self.computer_domain            = None

        # will be set to True if correlated to a computer object
        self.matched = False

        if LocalSession.SESSION_PTR in object.keys():
            if "reverse lookup failed" not in object[LocalSession.SESSION_PTR]:
                self.ptr_record = object[LocalSession.SESSION_PTR]
                self.computer_name = self.ptr_record.split('.')[0]
                self.computer_domain = '.'.join(self.ptr_record.split('.')[1:])
        
        if LocalSession.SESSION_USER in object.keys():
            self.username = object[LocalSession.SESSION_USER]

        if LocalSession.SESSION_COMPUTER_NAME in object.keys():
            self.computer_name = object[LocalSession.SESSION_COMPUTER_NAME]

        if LocalSession.SESSION_COMPUTER_DOMAIN in object.keys():
            self.computer_netbios_domain = object[LocalSession.SESSION_COMPUTER_DOMAIN]
    

    def should_import(self):
        # missing required attributes
        if self.username is None or self.ptr_record is None \
            and self.computer_name is None or (self.computer_domain is None and self.computer_netbios_domain is None):
            return False
        
        # do not import sessions if NetWkstaGetInfo failed
        fail = "NetWkstaGetInfo Failed;"
        if self.computer_name is not None and self.computer_netbios_domain is not None:
            if self.computer_name.startswith(fail) or self.computer_netbios_domain.startswith(fail):
                return False
        
        # do not import computer accounts
        if self.username.endswith('$'):
            return False
        
        # do not import anonymous sessions
        if self.username.upper() == 'ANONYMOUS LOGON':
            return False
        
        computer = self.ptr_record if self.ptr_record else self.computer_name
        logger.debug(f"NetSessionEnum session found for {ColorScheme.user}{self.username}[/] on {ColorScheme.computer}{computer}[/]", extra=OBJ_EXTRA_FMT)
        return True


    # so that a set can be used to keep a unique list of objects
    def __eq__(self, other):
        return (self.username, self.ptr_record, self.computer_name, self.computer_domain) == \
               (other.username, other.ptr_record, other.computer_name, other.computer_domain)
    

    # so that a set can be used to keep a unique list of objects
    def __hash__(self):
        return hash((self.username, self.ptr_record, self.computer_name, self.computer_domain))

    # for debugging
    def __repr__(self):
        return f"LocalSession(username={self.username}, ptr_record={self.ptr_record}, computer_name={self.computer_name}, computer_domain={self.computer_domain}, computer_netbios_domain={self.computer_netbios_domain})"


import re
import codecs

from bofhound.parsers import LdapSearchBofParser


class HavocParser(LdapSearchBofParser):

    @staticmethod
    def prep_file(file):
        with codecs.open(file, 'r', 'utf-8', errors='ignore') as f:
            contents = f.read()

        return re.sub(r'\[\d{2}\/\d{2}\/\d{4} \d{2}:\d{2}:\d{2}\] \[\+\] Received Output \[\d+ bytes\]:\n', '', contents)


from bloodhound.ad.utils import ADUtils

from .bloodhound_object import BloodHoundObject
from bofhound.logger import logger, OBJ_EXTRA_FMT, ColorScheme


class BloodHoundOU(BloodHoundObject):

    GUI_PROPERTIES = [
        'distinguishedname', 'whencreated',
        'domain', 'domainsid', 'name', 'highvalue', 'description',
        'blocksinheritance', 'isaclprotected'
    ]

    COMMON_PROPERTIES = [
    ]

    def __init__(self, object):
        super().__init__(object)

        self._entry_type = "OU"
        self.GPLinks = []
        self.ContainedBy = {}
        self.Properties["blocksinheritance"] = False
        
        if 'distinguishedname' in object.keys() and 'ou' in object.keys():
            self.Properties["domain"] = ADUtils.ldap2domain(object.get('distinguishedname').upper())
            self.Properties["name"] = f"{object.get('ou').upper()}@{self.Properties['domain']}"
            logger.debug(f"Reading OU object {ColorScheme.ou}{self.Properties['name']}[/]", extra=OBJ_EXTRA_FMT)

        if 'objectguid' in object.keys():
            self.ObjectIdentifier = object.get("objectguid").upper().upper()

        if 'ntsecuritydescriptor' in object.keys():
            self.RawAces = object['ntsecuritydescriptor']

        if 'description' in object.keys():
            self.Properties["description"] = object.get('description')

        if 'gplink' in object.keys():
            # [['DN1', 'GPLinkOptions1'], ['DN2', 'GPLinkOptions2'], ...]
            self.GPLinks = [link.upper()[:-1].split(';') for link in object.get('gplink').split('[LDAP//')][1:]

        if 'gpoptions' in object.keys():
            gpoptions = object.get('gpoptions')
            if gpoptions == '1':
                self.Properties["blocksinheritance"] = True

        self.Properties["highvalue"] = False

        self.Aces = []
        self.Links = []
        self.ChildObjects = []
        self.GPOChanges = {
            "AffectedComputers": [],
            "AffectedUsers": [],
            "DcomUsers": [],
            "LocalAdmins": [],
            "PSRemoteUsers": [],
            "RemoteDesktopUsers": []
        }
        self.IsDeleted = False
        self.IsACLProtected = False


    def to_json(self, properties_level):
        self.Properties['isaclprotected'] = self.IsACLProtected
        ou = super().to_json(properties_level)

        ou["ObjectIdentifier"] = self.ObjectIdentifier
        ou["ContainedBy"] = self.ContainedBy
        # The below is all unsupported as of now.
        ou["Aces"] = self.Aces
        ou["Links"] = self.Links
        ou["ChildObjects"] = self.ChildObjects

        self.GPOChanges["AffectedComputers"] = self.AffectedComputers
        self.GPOChanges["AffectedUsers"] = self.AffectedUsers
        ou["GPOChanges"] = self.GPOChanges

        ou["IsDeleted"] = self.IsDeleted
        ou["IsACLProtected"] = self.IsACLProtected

        return ou



from bloodhound.ad.utils import ADUtils
from .bloodhound_object import BloodHoundObject
from bofhound.logger import logger, OBJ_EXTRA_FMT, ColorScheme


class BloodHoundContainer(BloodHoundObject):

    GUI_PROPERTIES = [
        'domain', 'name', 'distinguishedname', 'domainsid', 'highvalue', 'isaclprotected'
    ]

    COMMON_PROPERTIES = [
    ]

    def __init__(self, object):
        super().__init__(object)

        self._entry_type = "Container"
        self.ContainedBy = {}
        self.Properties["blocksinheritance"] = False

        if 'objectguid' in object.keys():
            self.ObjectIdentifier = object.get("objectguid").upper().upper()
        
        if 'distinguishedname' in object.keys() and 'ou' in object.keys():
            self.Properties["domain"] = ADUtils.ldap2domain(object.get('distinguishedname').upper())
            self.Properties["name"] = f"{object.get('name').upper()}@{self.Properties['domain']}"
            logger.debug(f"Reading Container object {ColorScheme.ou}{self.Properties['name']}[/]", extra=OBJ_EXTRA_FMT)
        
        self.Properties['highvalue'] = False

        if 'ntsecuritydescriptor' in object.keys():
            self.RawAces = object['ntsecuritydescriptor']

        self.Properties["highvalue"] = False

        self.Aces = []
        self.ChildObjects = []
        self.IsDeleted = False
        self.IsACLProtected = False


    def to_json(self, properties_level):
        self.Properties['isaclprotected'] = self.IsACLProtected
        ou = super().to_json(properties_level)

        ou["ObjectIdentifier"] = self.ObjectIdentifier
        ou["ContainedBy"] = self.ContainedBy
        ou["Aces"] = self.Aces
        ou["ChildObjects"] = self.ChildObjects
        ou["IsDeleted"] = self.IsDeleted
        ou["IsACLProtected"] = self.IsACLProtected

        return ou



import re
import codecs
import json

from bofhound.logger import logger
from bofhound.parsers.generic_parser import GenericParser
from bofhound.parsers import LdapSearchBofParser


#
# Parses ldapsearch BOF objects from Outflank C2 JSON logfiles
#   Assumes that the BOF was registered as a command in OC2 named 'ldapserach'
#

class OutflankC2JsonParser(LdapSearchBofParser):
    BOFNAME =  'ldapsearch'
    

    @staticmethod
    def prep_file(file):
        with codecs.open(file, 'r', 'utf-8') as f:
            return f.read()


    #
    # Slightly modified from LdapSearchBofParser to account for
    #  needing only part of each JSON object, instead of the whole file
    #
    @staticmethod
    def parse_data(contents):
        parsed_objects = []
        current_object = None
        in_result_region = False
        previous_attr = None

        in_result_region = False

        lines = contents.splitlines()
        for line in lines:
            event_json = json.loads(line.split('UTC ', 1)[1])

            # we only care about task_resonse events
            if event_json['event_type'] != 'task_response':
                continue
            
            # within task_response events, we only care about tasks with the name 'ldapsearch'
            if event_json['task']['name'].lower() !=  OutflankC2JsonParser.BOFNAME:
                continue
            
            # now we have a block of ldapsearch data we can parse through for objects
            response_lines = event_json['task']['response'].splitlines()
            for response_line in response_lines:

                is_boundary_line = OutflankC2JsonParser._is_boundary_line(response_line)

                if (not in_result_region and
                    not is_boundary_line):
                    continue

                if (is_boundary_line
                    and is_boundary_line != OutflankC2JsonParser._COMPLETE_BOUNDARY_LINE):
                    while True:
                        try:
                            next_line = next(response_lines)[1]
                            remaining_length = OutflankC2JsonParser._is_boundary_line(next_line, is_boundary_line)

                            if remaining_length:
                                is_boundary_line = remaining_length
                                if is_boundary_line == OutflankC2JsonParser._COMPLETE_BOUNDARY_LINE:
                                    break
                        except:
                            # probably ran past the end of the iterable
                            break

                if (is_boundary_line):
                    if not in_result_region:
                        in_result_region = True
                    elif current_object is not None:
                        # self.store_object(current_object)
                        parsed_objects.append(current_object)
                    current_object = {}
                    continue
                elif re.match("^(R|r)etr(e|i)(e|i)ved \\d+ results?", response_line):
                    #self.store_object(current_object)
                    parsed_objects.append(current_object)
                    in_result_region = False
                    current_object = None
                    continue

                data = response_line.split(': ')

                try:
                    # If we previously encountered a control message, we're probably still in the old property
                    if len(data) == 1:
                        if previous_attr is not None:
                            value = current_object[previous_attr] + response_line
                    else:
                        data = response_line.split(':')
                        attr = data[0].strip().lower()
                        value = ''.join(data[1:]).strip()
                        previous_attr = attr

                    current_object[attr] = value

                except Exception as e:
                    logger.debug(f'Error - {str(e)}')

        return parsed_objects


    #
    # Get local groups, sessions, etc by feeding data to GenericParser class
    #
    @staticmethod
    def parse_local_objects(file):
        return GenericParser.parse_file(file, is_outflankc2=True)


from .netloggedon_bof import NetLoggedOnBofParser
from .netsession_bof import NetSessionBofParser
from .netlocalgroup_bof import NetLocalGroupBofParser
from .regsession_bof import RegSessionBofParser

__all_generic_parsers__ = [
    NetLoggedOnBofParser,
    NetSessionBofParser,
    NetLocalGroupBofParser,
    RegSessionBofParser
]


from bloodhound.ad.utils import ADUtils

from .bloodhound_object import BloodHoundObject
from bofhound.logger import logger, OBJ_EXTRA_FMT, ColorScheme
from bofhound.ad.helpers.cert_utils import PkiCertificateAuthorityFlags


class BloodHoundEnterpriseCA(BloodHoundObject):

    GUI_PROPERTIES = [
        'domain', 'name', 'distinguishedname', 'domainsid', 'isaclprotected',
        'description', 'whencreated', 'flags', 'caname', 'dnshostname', 'certthumbprint',
        'certname', 'certchain', 'hasbasicconstraints', 'basicconstraintpathlength',
        'casecuritycollected', 'enrollmentagentrestrictionscollected', 'isuserspecifiessanenabledcollected',
        'unresolvedpublishedtemplates'
    ]

    COMMON_PROPERTIES = [
    ]

    def __init__(self, object):
        super().__init__(object)

        self._entry_type = "EnterpriseCA"
        self.IsDeleted = False
        self.ContainedBy = {}
        self.IsACLProtected = False
        self.Properties['casecuritycollected'] = False
        self.Properties['enrollmentagentrestrictionscollected'] = False
        self.Properties['isuserspecifiessanenabledcollected'] = False
        self.Properties['unresolvedpublishedtemplates'] = []
        self.CARegistryData = None
        self.x509Certificate = None

        if 'objectguid' in object.keys():
            self.ObjectIdentifier = object.get("objectguid").upper()

        if 'distinguishedname' in object.keys():
            domain = ADUtils.ldap2domain(object.get('distinguishedname')).upper()
            self.Properties['domain'] = domain
            self.Properties['distinguishedname'] = object.get('distinguishedname').upper()

        if 'description' in object.keys():
            self.Properties['description'] = object.get('description')
        else:
            self.Properties['description'] = None

        if 'flags' in object.keys():
            int_flag = int(object.get("flags"))
            self.Properties['flags'] = ', '.join([member.name for member in PkiCertificateAuthorityFlags if member.value & int_flag == member.value])

        if 'name' in object.keys():
            self.Properties['caname'] = object.get('name')
            if 'domain' in self.Properties.keys():
                self.Properties['name'] = object.get('name').upper() + "@" + self.Properties['domain'].upper()

        if 'dnshostname' in object.keys():
            self.Properties['dnshostname'] = object.get('dnshostname')

        if 'cacertificate' in object.keys():
            self.parse_cacertificate(object)

        if 'ntsecuritydescriptor' in object.keys():
            self.RawAces = object['ntsecuritydescriptor']

        self.HostingComputer = None
        self.EnabledCertTemplates = []

        if 'certificatetemplates' in object.keys():
            self.CertTemplates = object.get('certificatetemplates').split(', ')
    

    def to_json(self, properties_level):
        self.Properties['isaclprotected'] = self.IsACLProtected
        data = super().to_json(properties_level)

        data["HostingComputer"] = self.HostingComputer
        data["CARegistryData"] = self.CARegistryData
        data["EnabledCertTemplates"] = self.EnabledCertTemplates
        data["Aces"] = self.Aces
        data["ObjectIdentifier"] = self.ObjectIdentifier
        data["IsDeleted"] = self.IsDeleted
        data["IsACLProtected"] = self.IsACLProtected
        data["ContainedBy"] = self.ContainedBy
        
        return data


import sys
import os
import logging
import typer
import glob
from syncer import sync

from bofhound.parsers import LdapSearchBofParser, Brc4LdapSentinelParser, HavocParser, \
    ParserType, OutflankC2JsonParser, MythicParser
from bofhound.writer import BloodHoundWriter
from bofhound.uploader import BloodHoundUploader
from bofhound.ad import ADDS
from bofhound.local import LocalBroker
from bofhound import console
from bofhound.ad.helpers import PropertiesLevel
from bofhound.logger import logger

app = typer.Typer(
    add_completion=False,
    rich_markup_mode="rich",
    context_settings={'help_option_names': ['-h', '--help']}
)

@app.command()
def main(
    input_files: str = typer.Option("/opt/cobaltstrike/logs", "--input", "-i", help="Directory or file containing logs of ldapsearch results"),
    output_folder: str = typer.Option(".", "--output", "-o", help="Location to export bloodhound files"),
    properties_level: PropertiesLevel = typer.Option(PropertiesLevel.Member.value, "--properties-level", "-p", case_sensitive=False, help='Change the verbosity of properties exported to JSON: Standard - Common BH properties | Member - Includes MemberOf and Member | All - Includes all properties'),
    parser: ParserType = typer.Option(ParserType.LdapsearchBof.value, "--parser", case_sensitive=False, help="Parser to use for log files. ldapsearch parser (default) supports ldapsearch BOF logs from Cobalt Strike and pyldapsearch logs"),
    debug: bool = typer.Option(False, "--debug", help="Enable debug output"),
    zip_files: bool = typer.Option(False, "--zip", "-z", help="Compress the JSON output files into a zip archive"),
    quiet: bool = typer.Option(False, "--quiet", "-q", help="Suppress banner"),
    mythic_server: str = typer.Option("127.0.0.1", "--mythic-server", help="IP or hostname of Mythic server to connect to", rich_help_panel="Mythic Options"),
    mythic_token: str = typer.Option(None, "--mythic-token", help="Mythic API token", rich_help_panel="Mythic Options"),
    bh_token_id: str = typer.Option(None, "--bh-token-id", help="BloodHound API token ID", rich_help_panel="BloodHound CE Options"),
    bh_token_key: str = typer.Option(None, "--bh-token-key", help="BloodHound API token key", rich_help_panel="BloodHound CE Options"),
    bh_server: str = typer.Option("http://127.0.0.1:8080", "--bh-server", help="BloodHound CE URL", rich_help_panel="BloodHound CE Options")):
    """
    Generate BloodHound compatible JSON from logs written by the ldapsearch BOF, pyldapsearch and specific C2 frameworks
    """

    if debug:
        logging.getLogger().setLevel(logging.DEBUG)
    else:
        logging.getLogger().setLevel(logging.INFO)

    if not quiet:
        banner()

     # default to Cobalt logfile naming format
    logfile_name_format = "beacon*.log"

    match parser:
        
        case ParserType.LdapsearchBof:
            logger.debug("Using ldapsearch parser")
            parser = LdapSearchBofParser
        
        case ParserType.BRC4:
            logger.debug("Using Brute Ratel parser")
            parser = Brc4LdapSentinelParser
            logfile_name_format = "b-*.log"
            if input_files == "/opt/cobaltstrike/logs":
                input_files = "/opt/bruteratel/logs"

        case ParserType.HAVOC:
            logger.debug("Using Havoc parser")
            parser = HavocParser
            logfile_name_format = "Console_*.log"
            if input_files == "/opt/cobaltstrike/logs":
                input_files = "/opt/havoc/data/loot"

        case ParserType.OUTFLANKC2:
            logger.debug("Using OutflankC2 parser")
            parser = OutflankC2JsonParser
            logfile_name_format = "*.json"

        case ParserType.MYTHIC:
            logger.debug("Using Mythic parser")
            parser = MythicParser()
            if mythic_token is None:
                logger.error("Mythic server and API token must be provided")
                sys.exit(-1)
            #
            # instead of iteraitng over log files on disk, we'll iterate over
            # Mythic callback objects
            #
            sync(parser.connect(mythic_server, mythic_token))
            cs_logs = sync(parser.collect_callbacks())

        
        case _:
            raise ValueError(f"Unknown parser type: {parser}")

    if os.path.isfile(input_files):
        cs_logs = [input_files]
        logger.debug(f"Log file explicitly provided {input_files}")
    elif os.path.isdir(input_files):
        # recurisively get a list of all .log files in the input directory, sorted by last modified time
        cs_logs = glob.glob(f"{input_files}/**/{logfile_name_format}", recursive=True)
        if len(cs_logs) == 0:
            # check for pyldapsearch/soapy logs
            cs_logs = glob.glob(f"{input_files}/*.log", recursive=True)

        cs_logs.sort(key=os.path.getmtime)

        if len(cs_logs) == 0:
            logger.error(f"No log files found in {input_files}!")
            return
        else:
            logger.info(f"Located {len(cs_logs)} beacon log files")
    else:
        if not isinstance(parser, MythicParser):
            logger.error(f"Could not find {input_files} on disk")
            sys.exit(-1)

    parsed_ldap_objects = []
    parsed_local_objects = []
    with console.status(f"", spinner="aesthetic") as status:
        for log in cs_logs:
            status.update(f" [bold] Parsing {log}")
            formatted_data = parser.prep_file(log)
            new_objects = parser.parse_data(formatted_data)
            
            # jank insert to reparse outflank logs for local data
            if parser == OutflankC2JsonParser:
                new_local_objects = parser.parse_local_objects(log)
            else:
                new_local_objects = parser.parse_local_objects(formatted_data)
            
            logger.debug(f"Parsed {log}")
            logger.debug(f"Found {len(new_objects)} objects in {log}")
            parsed_ldap_objects.extend(new_objects)
            parsed_local_objects.extend(new_local_objects)

    logger.info(f"Parsed {len(parsed_ldap_objects)} LDAP objects from {len(cs_logs)} log files")
    logger.info(f"Parsed {len(parsed_local_objects)} local group/session objects from {len(cs_logs)} log files")

    ad = ADDS()
    broker = LocalBroker()

    logger.info("Sorting parsed objects by type...")
    ad.import_objects(parsed_ldap_objects)
    broker.import_objects(parsed_local_objects, ad.DOMAIN_MAP.values())

    logger.info(f"Parsed {len(ad.users)} Users")
    logger.info(f"Parsed {len(ad.groups)} Groups")
    logger.info(f"Parsed {len(ad.computers)} Computers")
    logger.info(f"Parsed {len(ad.domains)} Domains")
    logger.info(f"Parsed {len(ad.trustaccounts)} Trust Accounts")
    logger.info(f"Parsed {len(ad.ous)} OUs")
    logger.info(f"Parsed {len(ad.containers)} Containers")
    logger.info(f"Parsed {len(ad.gpos)} GPOs")
    logger.info(f"Parsed {len(ad.enterprisecas)} Enterprise CAs")
    logger.info(f"Parsed {len(ad.aiacas)} AIA CAs")
    logger.info(f"Parsed {len(ad.rootcas)} Root CAs")
    logger.info(f"Parsed {len(ad.ntauthstores)} NTAuth Stores")
    logger.info(f"Parsed {len(ad.issuancepolicies)} Issuance Policies")
    logger.info(f"Parsed {len(ad.certtemplates)} Cert Templates")
    logger.info(f"Parsed {len(ad.schemas)} Schemas")
    logger.info(f"Parsed {len(ad.CROSSREF_MAP)} Referrals")
    logger.info(f"Parsed {len(ad.unknown_objects)} Unknown Objects")
    logger.info(f"Parsed {len(broker.sessions)} Sessions")
    logger.info(f"Parsed {len(broker.privileged_sessions)} Privileged Sessions")
    logger.info(f"Parsed {len(broker.registry_sessions)} Registry Sessions")
    logger.info(f"Parsed {len(broker.local_group_memberships)} Local Group Memberships")

    ad.process()
    ad.process_local_objects(broker)

    #
    # Write out the BloodHound JSON files
    #
    outfiles = BloodHoundWriter.write(
        output_folder,
        domains=ad.domains,
        computers=ad.computers,
        users=ad.users,
        groups=ad.groups,
        ous=ad.ous,
        containers=ad.containers,
        gpos=ad.gpos,
        enterprisecas=ad.enterprisecas,
        aiacas=ad.aiacas,
        rootcas=ad.rootcas,
        ntauthstores=ad.ntauthstores,
        issuancepolicies=ad.issuancepolicies,
        certtemplates = ad.certtemplates,
        properties_level=properties_level,
        zip_files=zip_files
    )

    #
    # Upload files to BloodHound CE
    #
    if bh_token_id and bh_token_key and bh_server:
        with console.status(f"", spinner="aesthetic") as status:
            status.update(f" [bold] Uploading files to BloodHound server...")
            uploader = BloodHoundUploader(bh_server, bh_token_id, bh_token_key)
            
            if not uploader.create_upload_job():
                return

            for file in outfiles:
                uploader.upload_file(file)

            uploader.close_upload_job()
        logger.info("Files uploaded to BloodHound server")


def banner():
    print('''
 _____________________________ __    __    ______    __    __   __   __   _______
|   _   /  /  __   / |   ____/|  |  |  |  /  __  \\  |  |  |  | |  \\ |  | |       \\
|  |_)  | |  |  |  | |  |__   |  |__|  | |  |  |  | |  |  |  | |   \\|  | |  .--.  |
|   _  <  |  |  |  | |   __|  |   __   | |  |  |  | |  |  |  | |  . `  | |  |  |  |
|  |_)  | |  `--'  | |  |     |  |  |  | |  `--'  | |  `--'  | |  |\\   | |  '--'  |
|______/   \\______/  |__|     |__|  |___\\_\\________\\_\\________\\|__| \\___\\|_________\\

                            << @coffeegist | @Tw1sm >>
    ''')


if __name__ == "__main__":
    app(prog_name="bofhound")



from .sharedparser import SharedParser


class NetSessionBofParser(SharedParser):
    START_BOUNDARY  = "---------------Session--------------"
    END_BOUNDARY    = "-------------End Session------------"
    OBJECT_TYPE     = "Session"

    def __init__(self):
        pass


    @staticmethod
    def is_start_boundary_line(line):
        return line.strip() == NetSessionBofParser.START_BOUNDARY
    

    @staticmethod
    def is_end_boundary_line(line):
        return line.strip() == NetSessionBofParser.END_BOUNDARY